// This is public domain software and comes with
// absolutely no warranty. Use of public domain software
// may vary between counties, but in general you are free
// to use and distribute this software for any purpose.


// Example: OSG using an OpenGL 3.1 context.
// The comment block at the end of the source describes building OSG
// for use with OpenGL 3.x.

#include <osgViewer/Viewer>
#include <osgDB/ReadFile>
#include <osgDB/WriteFile>
#include <osg/GraphicsContext>
#include <osg/Camera>
#include <osg/Viewport>
#include <osg/StateSet>
#include <osg/Program>
#include <osg/Shader>
#include <osg/DrawMeshTasks>
#include <osg/DrawMeshTasksIndirect>
#include <osg/MultiDrawMeshTasksIndirect>
#include <osg/MultiDrawMeshTasksIndirectCount>
#include <osgUtil/Optimizer>
#include <osg/TextureBuffer>
#include <osg/TriangleFUnctor>

#include <osgViewer/ViewerEventHandlers>

#include <osgGA/TrackballManipulator>
#include <osgGA/FlightManipulator>
#include <osgGA/DriveManipulator>
#include <osgGA/KeySwitchMatrixManipulator>
#include <osgGA/StateSetManipulator>
#include <osgGA/AnimationPathManipulator>
#include <osgGA/TerrainManipulator>
#include <osgGA/SphericalManipulator>

#include <osg/MatrixTransform>

#include <osgSim/LightPointNode>

#include <limits>

#define MESH_VERTEX_ARRAY_TEXTURE_UNIT 3
#define MESH_INFOS_ARRAY_TEXTURE_UNIT 1
#define MESH_INDEX_ARRAY_TEXTURE_UNIT 2
#define MESH_UVS_ARRAY_TEXTURE_UNIT 4

using namespace osg;

namespace osg
{

	class ProxyPrimitive : public PrimitiveSet
	{
	public:

		ProxyPrimitive() { drawMeshTasks = new osg::DrawMeshTasks(0, 1); }

		ProxyPrimitive(const ProxyPrimitive& pp, const CopyOp& copyop = CopyOp::SHALLOW_COPY) :
			PrimitiveSet(pp, copyop),
			drawMeshTasks(pp.drawMeshTasks) {
			drawMeshTasks = new osg::DrawMeshTasks(0, 1);
		}

		META_Object(osg, ProxyPrimitive)

			ref_ptr<DrawMeshTasks> drawMeshTasks;

		void draw(State& state, bool) const
		{
			drawMeshTasks->draw(state);
		}

		void accept(PrimitiveFunctor&) const {}
		void accept(PrimitiveIndexFunctor&) const {}

		unsigned int index(unsigned int) const { return 0; }
		unsigned int getNumIndices() const { return 0; }
		void offsetIndices(int) {}
	};

	REGISTER_OBJECT_WRAPPER(ProxyPrimitive,
		new osg::ProxyPrimitive,
		osg::ProxyPrimitive,
		"osg::Object osg::BufferData osg::PrimitiveSet osg::ProxyPrimitive")
	{
		{
			UPDATE_TO_VERSION_SCOPED(147)
				ADDED_ASSOCIATE("osg::BufferData")
		}
	}

}

struct ToMeshletsVisitor : public osg::NodeVisitor
{
	ToMeshletsVisitor()
		: osg::NodeVisitor(osg::NodeVisitor::TRAVERSE_ALL_CHILDREN)
	{

	}

	
	struct Triangulate
	{
		struct Meshlet
		{
			size_t						_startVertices = 0;
			size_t						_startIndices = 0;
			std::vector<uint64_t>		_vertices;
			std::vector<uint64_t>		_indices;
			size_t						_primitiveCount = 0;
			std::map<osg::Vec3, bool>	_uniqueVerticesMap;

			inline bool exists(const osg::Vec3& v)
			{
				std::map<osg::Vec3, bool>::iterator itr = _uniqueVerticesMap.find(v);

				if (itr != _uniqueVerticesMap.end()) return true;
				return false;
			}
		};

		std::vector<Meshlet>				_meshletsFromGeometry;
		size_t								_vxsStart = 0;
		size_t								_idxStart = 0;
		std::map<osg::Vec3, osg::Vec2>		_uvMap;
		osg::ref_ptr<osg::Vec2Array>		_uvs;
		osg::Vec3Array*						_vxs;
		osg::ref_ptr<osg::UIntArray>		_indices = new osg::UIntArray;
		std::map<osg::Vec3, size_t>			_vxsMap;
		bool								_done = false;

		inline void setUVMap(const std::map<osg::Vec3, osg::Vec2>& uvMap)
		{
			_uvMap = uvMap;
		}

		inline void setUVs(osg::Vec2Array* uvs)
		{
			_uvs = uvs;
		}
		inline void set(osg::Vec3Array* vertices)
		{
			_vxs = vertices;
		}


		inline void operator()(const osg::Vec3& v1, const osg::Vec3& v2, const osg::Vec3& v3)
		{
			if (!_meshletsFromGeometry.size())
			{
				_meshletsFromGeometry.push_back(Meshlet());
			}			
			if (_done) return;

			bool	unique_v1 = false;
			bool	unique_v2 = false;
			bool	unique_v3 = false;
			size_t	num_unique_verts = 0;			
			
			{
				Meshlet& meshlet = _meshletsFromGeometry.back();

				//if (!meshlet.exists(v1))
				{
					unique_v1 = true;
					++num_unique_verts;
				}
				//if (!meshlet.exists(v2))
				{
					unique_v2 = true;
					++num_unique_verts;
				}
				//if (!meshlet.exists(v3))
				{
					unique_v3 = true;
					++num_unique_verts;
				}
			}
			
			{
				Meshlet& meshlet = _meshletsFromGeometry.back();

				//if (num_unique_verts + meshlet._vertices.size() > 256)
				if (num_unique_verts + _vxs->size() > 30)
				{
					_vxsStart += meshlet._vertices.size();
					_idxStart += meshlet._indices.size();

					_meshletsFromGeometry.push_back(Meshlet());
					Meshlet& meshlet = _meshletsFromGeometry.back();

					meshlet._startVertices = _vxsStart;
					meshlet._startIndices = _idxStart;

					unique_v1 = true;					
					unique_v2 = true;					
					unique_v3 = true;					
				}
			}

			Meshlet& meshlet = _meshletsFromGeometry.back();
			
			std::map<osg::Vec3, size_t>::iterator itr = _vxsMap.find(v1);
			if (1)//(itr == _vxsMap.end())
			{
				_indices->push_back(_vxs->size());
				_vxsMap[v1] = _vxs->size();

				_vxs->push_back(v1);
				_uvs->push_back(_uvMap[v1]);
			}
			else
			{
				_indices->push_back(itr->second);
			}
			meshlet._indices.push_back(_indices->back());

			itr = _vxsMap.find(v2);
			if (1)//(itr == _vxsMap.end())
			{
				_indices->push_back(_vxs->size());
				_vxsMap[v2] = _vxs->size();

				_vxs->push_back(v2);
				_uvs->push_back(_uvMap[v2]);
			}
			else
			{
				_indices->push_back(itr->second);
			}
			meshlet._indices.push_back(_indices->back());

			itr = _vxsMap.find(v3);
			if (1)//(itr == _vxsMap.end())
			{
				_indices->push_back(_vxs->size());
				_vxsMap[v3] = _vxs->size();

				_vxs->push_back(v3);
				_uvs->push_back(_uvMap[v3]);
			}
			else
			{
				_indices->push_back(itr->second);
			}	
			meshlet._indices.push_back(_indices->back());

			++meshlet._primitiveCount;
	
			{
				Meshlet& meshlet = _meshletsFromGeometry.back();

				if (meshlet._primitiveCount + 1 == 512)
				{

					_vxsStart += meshlet._vertices.size();
					_idxStart += meshlet._indices.size();

					_meshletsFromGeometry.push_back(Meshlet());
					Meshlet& meshlet = _meshletsFromGeometry.back();

					meshlet._startVertices = _vxsStart;
					meshlet._startIndices = _idxStart;

					unique_v1 = true;
					unique_v2 = true;
					unique_v3 = true;
				}

				{
					Meshlet& meshlet = _meshletsFromGeometry.back();

					if (unique_v1) { meshlet._vertices.push_back(_vxsMap[v1]); meshlet._uniqueVerticesMap[v1] = true; }
					if (unique_v2) { meshlet._vertices.push_back(_vxsMap[v2]); meshlet._uniqueVerticesMap[v2] = true; }
					if (unique_v3) { meshlet._vertices.push_back(_vxsMap[v3]); meshlet._uniqueVerticesMap[v3] = true; }
				}
			}

		}		

	};

	virtual void apply(osg::Node& node)
	{
		osg::ref_ptr<osgSim::LightPointNode> lpn = dynamic_cast<osgSim::LightPointNode*>(&node);
		if (lpn.valid())
		{
			lpn->setNodeMask(0);
			return;
		}
		traverse(node);
	}		

	virtual void apply(osg::Geode& node)
	{
		for (size_t i = 0; i < node.getNumDrawables(); ++i)
		{
			osg::ref_ptr<osg::Geometry> g = node.getDrawable(i)->asGeometry();
			if (g.valid())
			{							
				if (g->getNumPrimitiveSets() == 0) continue;
				if (g->getPrimitiveSet(0)->getMode() == GL_POINTS) continue;
				if (g->getPrimitiveSet(0)->getMode() == GL_POLYGON) continue;
								
				osg::ref_ptr<osg::Vec3Array> vxs = new osg::Vec3Array;
				osg::ref_ptr<osg::Vec2Array> muvs = new osg::Vec2Array;

				std::map<osg::Vec3, osg::Vec2> uv_map;

				osg::ref_ptr<osg::Vec3Array> vs = dynamic_cast<osg::Vec3Array*>(g->getVertexArray());
				osg::ref_ptr<osg::Vec2Array> uvs = dynamic_cast<osg::Vec2Array*>(g->getTexCoordArray(0));
				if (uvs.valid() && vs.valid())
				{
					for (size_t i = 0; i < uvs->size(); ++i)
					{
						uv_map[vs->at(i)] = uvs->at(i);
					}
				}
				
				osg::TriangleFunctor<Triangulate> t;
				t.set(vxs);
				t.setUVs(muvs);
				t.setUVMap(uv_map);

				g->accept(t);	

				if (t._vxs->size() < 3) continue;
				if (t._indices->size() < 3) continue;
				
				
				

				osg::ref_ptr<FloatArray> MeshletInfosArray = new FloatArray(t._meshletsFromGeometry.size());
				for (auto i = 0; i < t._meshletsFromGeometry.size(); ++i)
				{
					(*MeshletInfosArray)[i] = t._meshletsFromGeometry.at(i)._startVertices;
					//(*MeshletInfosArray)[i][1] = t._meshletsFromGeometry.at(i)._vertices.size();
					//(*MeshletInfosArray)[i][2] = t._meshletsFromGeometry.at(i)._startIndices;
					//(*MeshletInfosArray)[i][3] = t._meshletsFromGeometry.at(i)._indices.size();		
				}

				osg::ref_ptr<osg::Geometry> geometry = new osg::Geometry;
				geometry->setUseDisplayList(false);

				{
					osg::BoundingBox bb;

					size_t numVertices = vxs->size();
					osg::ref_ptr<osg::Vec4Array> osgVertices = new osg::Vec4Array(numVertices);
					for (auto j = 0; j < numVertices; ++j) {    // fill array

						(*osgVertices)[j] = osg::Vec4(vxs->at(j), 1);

						bb.expandBy(vxs->at(j));
					}

					//osg::ref_ptr<osg::FloatArray> osgIndices = new osg::FloatArray(t._indices.size());
					//for (auto j = 0; j < t._meshletsFromGeometry.at(i)._indices.size(); ++j) {    // fill array
					//	(*osgIndices)[j] = t._indices->at(j + t._meshletsFromGeometry.at(i)._startIndices);
					//}

					float r = std::rand()* 1.0f / RAND_MAX;
					float g = std::rand()* 1.0f / RAND_MAX;
					osg::ref_ptr<osg::Vec4Array> osguvs = new osg::Vec4Array(t._uvs->size());
					for (auto j = 0; j < t._uvs->size(); ++j) {    // fill array

						(*osguvs)[j] = osg::Vec4(t._uvs->at(j).x(), t._uvs->at(j).y(), r, g);
					}

					// Create Buffer and Texture
					osg::ref_ptr<osg::VertexBufferObject> verticesBufferObject = new osg::VertexBufferObject();
					verticesBufferObject->setUsage(GL_DYNAMIC_READ_ARB);
					osgVertices->setBufferObject(verticesBufferObject.get());
					osg::ref_ptr<osg::TextureBuffer> vertexTexture = new osg::TextureBuffer(osgVertices.get());
					vertexTexture->setInternalFormat(GL_RGBA32F_ARB);
					vertexTexture->setInternalFormatMode(Texture::USE_USER_DEFINED_FORMAT);
					geometry->getOrCreateStateSet()->setTextureAttributeAndModes(MESH_VERTEX_ARRAY_TEXTURE_UNIT, vertexTexture.get());

					// Create Buffer and Texture
					//osg::ref_ptr<osg::VertexBufferObject> indicesBufferObject = new osg::VertexBufferObject();
					//indicesBufferObject->setUsage(GL_DYNAMIC_READ_ARB);
					//osgIndices->setBufferObject(indicesBufferObject.get());
					//osg::ref_ptr<osg::TextureBuffer> indexTexture = new osg::TextureBuffer(osgIndices.get());
					//indexTexture->setInternalFormat(GL_LUMINANCE32F_ARB);
					//indexTexture->setInternalFormatMode(Texture::USE_USER_DEFINED_FORMAT);
					//geometry->getOrCreateStateSet()->setTextureAttributeAndModes(MESH_INDEX_ARRAY_TEXTURE_UNIT, indexTexture.get());

					// Create Buffer and Texture
					osg::ref_ptr<osg::VertexBufferObject> uvsBufferObject = new osg::VertexBufferObject();
					uvsBufferObject->setUsage(GL_DYNAMIC_READ_ARB);
					osguvs->setBufferObject(uvsBufferObject.get());
					osg::ref_ptr<osg::TextureBuffer> uvsTexture = new osg::TextureBuffer(osguvs.get());
					uvsTexture->setInternalFormat(GL_RGBA32F_ARB);
					uvsTexture->setInternalFormatMode(Texture::USE_USER_DEFINED_FORMAT);
					geometry->getOrCreateStateSet()->setTextureAttributeAndModes(MESH_UVS_ARRAY_TEXTURE_UNIT, uvsTexture.get());

					geometry->setComputeBoundingBoxCallback(new MyCB(bb));

					meshlets.push_back(geometry);
				}

				osg::ref_ptr<osg::ProxyPrimitive> proxyPrimitive = new osg::ProxyPrimitive;
				proxyPrimitive->drawMeshTasks = new osg::DrawMeshTasks(0, t._meshletsFromGeometry.size());
				geometry->addPrimitiveSet(proxyPrimitive.get());

				
				//// Create Buffer and Texture
				osg::ref_ptr<osg::VertexBufferObject> meshInfosBufferObject = new osg::VertexBufferObject();
				meshInfosBufferObject->setUsage(GL_DYNAMIC_READ_ARB);
				MeshletInfosArray->setBufferObject(meshInfosBufferObject.get());
				osg::ref_ptr<osg::TextureBuffer> meshInfosTexture = new osg::TextureBuffer(MeshletInfosArray.get());
				meshInfosTexture->setInternalFormat(GL_LUMINANCE32F_ARB);
				meshInfosTexture->setInternalFormatMode(Texture::USE_USER_DEFINED_FORMAT);
				geometry->getOrCreateStateSet()->setTextureAttributeAndModes(MESH_INFOS_ARRAY_TEXTURE_UNIT, meshInfosTexture.get());
				
								
			}
		}		

		osg::ref_ptr<osg::Geode> geode = new osg::Geode;

		for (size_t i = 0; i < meshlets.size(); ++i)
		{
			geode->addDrawable(meshlets.at(i));
		}
		meshlets.clear();

		geode->setStateSet(node.getStateSet());

		node.getParent(0)->replaceChild(&node, geode);
		
	}

	std::vector<osg::ref_ptr<osg::Geometry>> meshlets;

	struct MyCB : public osg::Drawable::ComputeBoundingBoxCallback
	{
		MyCB(const osg::BoundingBox& bb)
			: _bb(bb)
		{

		}

		virtual BoundingBox computeBound(const osg::Drawable&) const
		{			
			return _bb;
		}

		osg::BoundingBox _bb;
	};	
};

static const char* meshSource = \
	"#version 450 \n"
	"#extension GL_NV_mesh_shader : enable\n"
	"layout(local_size_x = 32) in;"
	"layout(max_vertices = 64) out;"
	"layout(max_primitives = 126) out;"
	"layout(triangles) out;"
	"uniform samplerBuffer vertices;"
	"uniform samplerBuffer meshletsInfo;"
	//"uniform samplerBuffer indices;"
	"uniform samplerBuffer uvs;"
	"uniform mat4 osg_ModelViewProjectionMatrix;"
	"layout (location = 0) out PerVertexData"
	"{"
	"  vec2 uv;"
	"  vec2 rg;"
	"} v_out[];"
	"void main()"
	"{"
		"uint mi = gl_WorkGroupID.x;"
		"vec4 u = texelFetch(meshletsInfo, int(mi));"
		"uint start = uint(u.x);"
		//"uint count = uint(u.y);"
		//"uint idx_start = uint(u.z);"
		"uint idx_count = textureSize(vertices);"
		"uint id = gl_LocalInvocationID.x;"
		//"for (uint i = 0; i < count; ++i)"
		"{"
			"vec4 t = texelFetch(vertices, int(start+id));"
			"gl_MeshVerticesNV[id].gl_Position = osg_ModelViewProjectionMatrix* t;"
			"vec4 uvrg = texelFetch(uvs, int(start+id));"
			"v_out[id].uv = uvrg.st;" // uv
			"v_out[id].rg = uvrg.zw;" // color
		"}"	
		//"for (uint i = 0; i < idx_count; ++i)"
		"{"
			//"vec4 idx = texelFetch(indices,int(id));"
			"gl_PrimitiveIndicesNV[id] = uint(id);"
		"}"
		"if (id == 0)"
			"gl_PrimitiveCountNV = 10;"
	"}";


static const char* fragmentSource = \
	"#version 450 \n"
	"#extension GL_NV_fragment_shader_barycentric : enable\n"
	"out vec4 color;"
	"uniform sampler2D baseTexture;"
	"in PerVertexData"
	"{"
	"  vec2 uv;"
	"  vec2 rg;"
	"} uvrg;"
	"void main()"
	"{"
		"color = texture2D(baseTexture, uvrg.uv) * vec4(uvrg.rg.r,uvrg.rg.g,1,1);"	
	"}";


class DatabaseReadCallback : public osgDB::Registry::ReadFileCallback
{
public:
	DatabaseReadCallback()
	{

	}

	virtual osgDB::ReaderWriter::ReadResult readNode(const std::string& filename, const osgDB::Options* options)
	{
		osgDB::ReaderWriter::ReadResult result;

		result = osgDB::Registry::instance()->readNodeImplementation(filename, options);
		if (result.getNode())
		{
			if (filename.find("master.ive"))
			{
				osg::ref_ptr<osg::Shader> vShader = new osg::Shader(osg::Shader::MESH, meshSource);
				osg::ref_ptr<osg::Shader> fShader = new osg::Shader(osg::Shader::FRAGMENT, fragmentSource);

				osg::ref_ptr<osg::Program> program = new osg::Program;
				program->addShader(vShader.get());
				program->addShader(fShader.get());

				osg::ref_ptr<osg::Group> group = new osg::Group;
				result.getNode()->getOrCreateStateSet()->setAttribute(program.get());

				result.getNode()->getOrCreateStateSet()->getOrCreateUniform("vertices", osg::Uniform::SAMPLER_BUFFER)->set((int)MESH_VERTEX_ARRAY_TEXTURE_UNIT);
				result.getNode()->getOrCreateStateSet()->getOrCreateUniform("meshletsInfo", osg::Uniform::SAMPLER_BUFFER)->set((int)MESH_INFOS_ARRAY_TEXTURE_UNIT);
				result.getNode()->getOrCreateStateSet()->getOrCreateUniform("indices", osg::Uniform::SAMPLER_BUFFER)->set((int)MESH_INDEX_ARRAY_TEXTURE_UNIT);
				result.getNode()->getOrCreateStateSet()->getOrCreateUniform("uvs", osg::Uniform::SAMPLER_BUFFER)->set((int)MESH_UVS_ARRAY_TEXTURE_UNIT);
			}

			ToMeshletsVisitor nv;
			result.getNode()->accept(nv);

			//osg::ref_ptr<osgDB::ReaderWriter::Options> options = new osgDB::ReaderWriter::Options("WriteImageHint=IncludeData");
			//osgDB::writeNodeFile(*result.getNode(), filename + ".osgb");
		}

		return result;
	}
};


int main( int argc, char** argv )
{
    osg::ArgumentParser arguments( &argc, argv );

	osgDB::Registry::instance()->setReadFileCallback(new DatabaseReadCallback);


    // group->addChild( new osg::DrawMeshTasksIndirect(0) ); // will require a buffer to be bound.
    // group->addChild( new osg::MultiDrawMeshTasksIndirect(0, 0, 0) ); // will require a buffer to be bound.
    // group->addChild( new osg::MultiDrawMeshTasksIndirectCount(0, 0, 0, 0) ); // will require a buffer to be bound.
    

    osgViewer::Viewer viewer(arguments);

	// add the state manipulator
	viewer.addEventHandler(new osgGA::StateSetManipulator(viewer.getCamera()->getOrCreateStateSet()));

	// add the thread model handler
	viewer.addEventHandler(new osgViewer::ThreadingHandler);

	// add the window size toggle handler
	viewer.addEventHandler(new osgViewer::WindowSizeHandler);

	// add the stats handler
	viewer.addEventHandler(new osgViewer::StatsHandler);

	// add the help handler
	viewer.addEventHandler(new osgViewer::HelpHandler(arguments.getApplicationUsage()));

	// add the record camera path handler
	viewer.addEventHandler(new osgViewer::RecordCameraPathHandler);

	// add the LOD Scale handler
	viewer.addEventHandler(new osgViewer::LODScaleHandler);

	// add the screen capture handler
	viewer.addEventHandler(new osgViewer::ScreenCaptureHandler);

	osg::ref_ptr<osg::Node> model = osgDB::readNodeFiles(arguments);

	

    viewer.setSceneData( model );
	viewer.realize();

	osg::setNotifyLevel(osg::NOTICE);

    // for non GL3/GL4 and non GLES2 platforms we need enable the osg_ uniforms that the shaders will use,
    // you don't need thse two lines on GL3/GL4 and GLES2 specific builds as these will be enable by default.
	viewer.getCamera()->getGraphicsContext()->getState()->setUseModelViewAndProjectionUniforms(true);
	viewer.getCamera()->getGraphicsContext()->getState()->setUseVertexAttributeAliasing(true);

    return( viewer.run() );
}
